# FindDupFiles.py
# Copyright 2004 Matt Heinzen

# This file is part of DupeFinder 1.0.
# 
# DupeFinder is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# DupeFinder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with File Duplicate Finder; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import sys, os
import string, re


# Returns a list of file extensions formatted for findDups()
# Extensions are separated, converted to lowercase, and stripped of all
# separator and wildcard characters
def fileExtsFromString(fileExtStr = ""):
	exts = re.split("\W+", string.lower(fileExtStr))
	if len(exts) == 1 and len(exts[0]) == 0:
		exts = []
	return exts

# Returns a list of identical file groups for all files under the given starting
# directory and with one of the given file extensions (or any filename if no
# extensions are supplied)
# Each file group contains a tuple of three items: the file size, the md5sum
# string, and a list of complete pathnames for all identical files
def findDups(rootDir, fileExts = []):

	# SIZE is file size in bytes
	# fileDict[ SIZE ][0] is the array of full file paths
	# fileDict[ SIZE ][1] is hash MD5SUM => array of full file paths
	# e.g. fileDict[ SIZE ][1][ MD5SUM ][2] is the 3rd file found with the
	# given SIZE and calculated MD5SUM
	fileDict = {}

	# store current directory
	currPath = os.getcwd() + os.sep

	# expand any ~/ at beginning of path to actual path, # e.g. /home/username/
	rootDir = re.sub("^~" + os.sep, os.environ['HOME'] + os.sep, rootDir)

	# expand root search directory path (does nothing if given path is already a
	# complete, rather than local, path)
	rootPath = os.path.join(currPath, rootDir)

	# verify the existence of the specified directory
	if not os.path.isdir(rootPath):
		print "ERROR: specified directory does not exist or not available"
		return None

	# build a regular expression to determine if a filename matches any of the
	# provided file extensions, or any file if no extensions are provided
	# the pattern will apply to lowercase filenames only
	if len(fileExts) == 0:
		extRE = re.compile("")
	else:
		pattern = "\.("
		for ext in fileExts:
			pattern = pattern + ext + "|"
		pattern = pattern[0:-1] + ")$"
		extRE = re.compile(pattern)

	# build a dictionary counting the number of files of every possible size and
	# matching the given extensions in the given directory tree
	os.path.walk(rootPath, addFilesToSizeDict, (fileDict, extRE))

	# remove all filesize keys with only one matching file
	# add to all other values with empty dictionaries (which will map md5sums
	# to lists of filepaths)
	for size in fileDict.keys():
		if len(fileDict[size][0]) == 1: del fileDict[size]
		else: fileDict[size].append({})

	# build mini-dictionaries under each file size for all files with matching
	# md5sums
	buildMD5SumDict(fileDict)

	# change back to the original directory
	os.chdir(currPath)

	# create the list of identical file groups from the file dictionary
	fileGroups = []
	for size in fileDict.keys():
		for md5sum in fileDict[size][1].keys():
			if len(fileDict[size][1][md5sum]) > 1:
				group = (size, md5sum, fileDict[size][1][md5sum])
				fileGroups.append(group)

	return fileGroups


# Adds files to a list in a dictionary keyed by file size
def addFilesToSizeDict(fileData, dirName, filenames):

	# fileData[0] is fileDict from FindDups()
	# fileData[1] is fileExts from FindDups()
	# NOTE: fileData[1] will likely change to a compiled regular expression for
	# detecting any of the supplied extensions

	# add every actual file to the list for its size
	# TODO: check extension
	for fname in filenames:
		if re.search(fileData[1], string.lower(fname)) != None:
			# get the full path for the file (this prevents any errors based on
			# local paths or assumptions about the current working directory)
			path = os.path.join(dirName, fname)
			if not os.path.isdir(path):
				# get the file size
				size = os.path.getsize(path)

				# create a new list for previously unseen file sizes
				# the list will be initialized with an empty list for storing the
				# pathnames of files with that size
				if not fileData[0].has_key(size):
					fileData[0][size] = [[]]

				# store the pathname under the file size
				fileData[0][size][0].append(path)


# Identifies identical files using MD5 sums in a dictionary of files already
# grouped by size
def buildMD5SumDict(fileDict):

	# NOTE: consider using the built-in md5 module for portability

	for size in fileDict.keys():
		for filepath in fileDict[size][0]:
			# calculate md5sum using standard UNIX shell utility
			md5sum = os.popen('md5sum "' + filepath + '"').read()[:32]

			# create a new list for previously unseen md5sums under the current
			# file size
			# the list will be initialized with store the pathnames of
			# identical files (same size and md5sum)
			if not fileDict[size][1].has_key(md5sum):
				fileDict[size][1][md5sum] = []

			# store the pathname under the file size and md5sum combo
			fileDict[size][1][md5sum].append(filepath)



# output all duplicates STDOUT
# ARG 1: top-level directory path
# ARG 2..N: filename extension (e.g. mp3)
if __name__ == "__main__":
	# print USAGE and exit if no arguments supplied
	if len(sys.argv) < 2:
		print "USAGE: FindDupFiles.py <ROOT SEARCH DIR> [FILE EXT 1] " \
			"[FILE EXT 2] ..."
		print "File names are not case sensitive, e.g. 'txt'matches " \
			"'*.txt', '*.TXT' and '*.tXt'"
		sys.exit()

	# make sure directory has no separator character at end
	# leave any separator at beginning, as this distinguishes absolute paths
	# from relative paths
	rootDir = sys.argv[1]
	if rootDir[-1] == os.sep: rootDir = rootDir[0:-1]

	# extract the valid file extensions to check for
	fileExts = fileExtsFromString(string.join(sys.argv[2:], " "))

	# find all duplicate files
	dups = findDups(rootDir, fileExts)

	# print all duplicate groupss (in no particular order) to STDOUT
	if dups != None:
		for group in dups:
			print "Size = " + str(group[0]) + ", MD5Sum = " + group[1]
			for path in group[2]:
				print path
			print


